'''This QpropertyJsRef module provides JS reference for a given C++ class' Q_PROPERTY statements'''

class QpropertyJsRef:
	'''This class acts as a storage unit for a class JS reference'''


	def __init__(self):

		self.name_space = ""
		self.class_name = ""
		# The C++ statement extracted from the code
		self.statement = ""
		# List of lines making the description
		self.description = [ ]


	def __str__(self):

		return self.get_as_text("  ", 1)


	def get_description_as_text(self, tab, offset):

		lead = offset * tab

		output_text = ""

		for line in self.description:

			output_text = f"{output_text}{lead}{line}\n"

		# print(f"Returning property description:\n{output_text}")

		return output_text


	def print_description(self, tab, offset):

		output_text = self.get_description_as_text(tab, offset)
		return output_text


	def get_as_text(self, tab, offset):

		lead = offset * tab

		output_text = f"{lead}{self.statement}\n"

		if len(self.description):

			indent_offset = offset + 1

			# lead = indent_offset * tab
			# output_text = f"{output_text}{lead}Description:\n"
			# indent_offset += 1

			output_text = f"{output_text}{self.get_description_as_text(tab, indent_offset)}\n"

		# print(f"Returning QpropertyJsRef as text:\n{output_text}")

		return output_text


	def print(self, tab, offset):

		output_text = self.get_as_text(tab, offset)

		print(output_text)







'''This ClassJsRef module provides JS reference for a given C++ class'''

class ClassJsRef:
	'''This class acts as a storage unit for a class JS reference'''
	def __init__(self):

		self.name_space = ""
		self.class_name = ""
		# List of lines making the description
		self.description = [ ]
		# List of QpropertyJsRef instances
		self.properties = [ ]
		# List of QinvokableJsRef instances
		self.invokables = [ ]
		self.valid = False


	def __str__(self):
		return self.get_as_text("  ", 1)


	def get_description_as_text(self, tab, offset):

		lead = offset * tab

		# print(f"The description: {self.description}\n")
		# print(f"The description has {len(self.description)} items.")

		output_text = ""

		for line in self.description:

			# print(f"Appending class description line: '{line}")

			output_text = f"{output_text}{lead}{line}"

			# print(f"Now output_text is:\n{output_text}")

		return output_text


	def print_description(self, tab, offset):

		print(f"{self.get_description_as_text(tab, offset)}")


	def get_properties_as_text(self, tab, offset):

		output_text = ""

		for property  in self.properties:
			output_text = f"{output_text}{property.get_as_text(tab, offset)}"

		# print(f"Returning properties as text: {output_text}")

		return output_text


	def print_properties(self, tab, offset):

		print(f"{self.get_properties_as_text(tab, offset)}")


	def get_invokables_as_text(self, tab, offset):

		output_text = ""

		for invokable  in self.invokables:
			output_text = f"{output_text}{invokable.get_as_text(tab, offset)}"

		return output_text


	def print_invokables(self, tab, offset):

		print(f"{self.get_invokables_as_text(tab, offset)}")


	def get_as_text(self, tab, offset):

		lead = offset * tab

		output_text = f"{lead}{self.name_space} {self.class_name}:\n\n"

		offset += 1
		lead = offset * tab

		if len(self.description):

			output_text = f"{output_text}{lead}Description:\n\n"

			indented_offset = offset + 1

			output_text = f"{output_text}{self.get_description_as_text(tab, indented_offset)}\n"

		if len(self.properties):

			output_text = f"{output_text}{lead}Properties: {len(self.properties)}\n\n"

			indented_offset = offset + 1

			output_text = f"{output_text}{self.get_properties_as_text(tab, indented_offset)}\n"

			#print(f"Got {output_text} for properties text.")

		if len(self.invokables):

			output_text = f"{output_text}{lead}Invokables: {len(self.invokables)}\n\n"

			indented_offset = offset + 1

			output_text = f"{output_text}{self.get_invokables_as_text(tab, indented_offset)}\n"

			# print(f"Got {output_text} for invokables text.")

		return output_text


	def print(self, tab, offset):

		print(f"{self.get_as_text(tab, offset)}")








#!/usr/bin/env python3

import os
import sys
import argparse
import re
import shutil
import subprocess
import shlex
import copy

import jsExtractFunctions as jsef;
import ClassJsRef as ClassJsRef;


# The main directory will be the top source directory: development
mainDir = os.path.normpath(os.getcwd());
# print("mainDir:" + mainDir + "\n");

parser = argparse.ArgumentParser(description='''Script to extract
JavaScript-related reference text from source code files.''')

# parser.add_argument('-i', '--infiles', metavar='input_file', nargs='+',
# 										required = False, help='Source files to process.')

parser.add_argument('-t', '--tab', metavar='tabulator', required = False, help='String representing a tabulation key hit')

parser.add_argument('-o', '--outfile', metavar='output file', nargs='?', required = False, help='File name to output the results to (or the console if not specified.')

parser.add_argument('-l', '--listfile', metavar='file-list file', nargs='?', required = False, help='Name of the file that contains the list of files to process.')

parser.add_argument('-d', '--dir', metavar='dir', nargs='?', required = False, help='Directory where the listing of the JS reference-containing files should occur.')

group = parser.add_mutually_exclusive_group(required=True)

group.add_argument('--extract',
                    choices=['class', 'enum'],
                    help='Extract reference for specific type. Provide a list of files to be processed with the --infiles argument.')

group.add_argument('--list-files',
                    choices=['class', 'enum'],
                    help='List files having reference for specific type.')

args = parser.parse_args()

if args.tab:
	if args.tab == "\\t":
		tab = "\t"
	else:
		tab = args.tab
else:
	#tab = "\n"
	tab = "  "

# if args.infiles:
# 	print(f"Input files: {args.infiles}")
#
# if args.outfile:
# 	print(f"Output file: {args.outfile}")
#
# if args.extract:
# 	print(f"Extracting {args.extract}")
#
# if args.list:
# 	print(f"Listing {args.list}")


if args.listfile:
	with open(args.listfile, 'r') as f:
		content = f.read()
		# print(f"The contents of the file: {content}")
		args.infiles = content


if args.list_files:

	if args.dir:
		root_dir = args.dir
	else:
		root_dir = os.getcwd()

	if args.list_files == "class":
		file_list = jsef.listAllJsRefFiles(root_dir,
			"BEGIN CLASS JS REFERENCE")

	elif args.list_files == "enum":

		file_list = jsef.listAllJsRefFiles(root_dir,
			"BEGIN ENUM JS REFERENCE")

	# print(f"The listed files found in {root_dir}:\n {file_list}")

	# Now generate a list of files that is injectable to this same program
	# as the --infiles parameter.

	all_files_in_a_string =""

	for file in file_list:
		if file is not None:
			# print(f"{file} ")
			sys.stdout.write(file + ' ')

			#all_files_in_a_string = all_files_in_a_string + f" {file}"

	#print(all_files_in_a_string)


if args.extract:

	if not args.infiles:
		print("Provide a list of header file paths to process or the name of a file that contains that list.", file = sys.stderr)
		exit(1)

	# This is a list of ClassJsRef instances.
	all_class_js_ref_results = [ ]

	if args.extract == "class":

		# args.infiles is a list
		#print(type(args.infiles))

		file_list = args.infiles.split(' ')

		for file in file_list:

			# Now file is a string corresponding to a single file name.
			#print(type(file))
			# print(f"Current file: {file}")

			if file is not None and len(file) > 0:
				# simplified_file = re.sub(r'\s$', '', file)
				# print(f"\nNow processing file '{file}'\n\n")

				# The returned object is a list of ClassJsRef instances
				# in case more than one class is JS reference documented.
				class_js_ref_results = jsef.extract_class_js_reference(file)

				# print(f"Classes parsed in {file}:\n")

				for class_js_ref in class_js_ref_results:
					print(f"{str(class_js_ref)}\n")

					# There might be nothing in there as Q_PROPERTY or Q_INVOKABLE stuff
					if not len(class_js_ref.properties) and not len(class_js_ref.invokables):
						# print("For id {results[0]}n there were no Q_PROPERTY or Q_INVOKABLE")
						pass
					else:
						all_class_js_ref_results.append(class_js_ref)


		# print("All the results:\n")
		# all_class_js_ref_results is a list of  [ id, properties, invokables]
		# with id a tuple(name_space, class_name)
		# with properties a [ property_string ]
		# with invokables a [ invokable_string ]

		offset = 0

		#print(f"Text{lead}four tabs")
		#print(f"Text{tab}one tabs")

		if args.outfile:

			# print(f"Now outputting the extracted JS reference text to:\n{args.outfile}\n")

			from pathlib import Path
			file_path = Path(args.outfile)

			file_handle = file_path.open('w', encoding='utf-8')

			for class_js_ref_result in all_class_js_ref_results:

				file_handle.write(class_js_ref_result.get_as_text(tab, offset))

		else:

			# print(f"Now outputting the extracted JS reference text to terminal\n")

			for class_js_ref_result in all_class_js_ref_results:

				class_js_ref_result.print(tab, offset)



import os
import sys
import argparse
import re
import shutil
import subprocess
import shlex
import copy
import itertools
from pathlib import Path

import ClassJsRef as ClassJsRef
import QinvokableJsRef as QinvokableJsRef
import QpropertyJsRef as QpropertyJsRef

# Matches ... namespace: ... line
name_space_regex_string = r"^\s*\*?\s*namespace\s*:\s*([^\s.]+)\s*$"
name_space_regex = re.compile(name_space_regex_string)

# Matches ... class name: ... line
class_name_regex_string = r"^\s*\*?\s*class name\s*:\s*([^\s.]+)\s*$"
class_name_regex = re.compile(class_name_regex_string)

# Matches property type and property name
property_regex_string = r"^\s*Q_PROPERTY\(\s*([^\s.]+)\s+(\*?\s*[^\s?]+).*$"
property_regex = re.compile(property_regex_string)

# Matches everything after Q_INVOKABLE
invokable_regex_string = r"^\s*Q_INVOKABLE\s*(.*)$"
invokable_regex = re.compile(invokable_regex_string)


def parse_q_property_block(opening_line, file_handle):

	# opening_line can start these two ways:

	# 1. /*$ JS PROP REF The .... .... */
	# 2. Q_PROPERTY(Type variable WRITE write READ read)

	q_property_js_ref = QpropertyJsRef.QpropertyJsRef()

	ref_line_found = False

	js_ref_lines = [ ]
	statement_lines = [ ]

	opening_line = re.sub(r'^\s*', '', opening_line)
	opening_line = re.sub(r'\s*\n$', '', opening_line)


	if "$ JS PROP REF" in opening_line:

		# print(f"JS PROP REF: {opening_line}")

		ref_line_found = True

		js_ref_lines.append(opening_line)

		if "*/" not in opening_line:

			# print("*/ not in opening_line, multiline text")

			while(True):

				line = file_handle.readline()
				line = line.replace('\n', '')
				# In multiline, there will be chars at the beginning
				# and maybe at the end. Remove them because we need
				# to join the lines laters with ' ' and we do not want
				# multiple spaces in the finished text line.
				line = re.sub(r'^\s*', '', line)
				line = re.sub(r'\s*$', '', line)

				# print(f"Read follow-up JS PROP REF line '{line}'")
				js_ref_lines.append(line)

				if "*/" in line:

					# We have finished the JS REF block for the property to come.
					break;


		# Join the lines into a single string
		js_ref_as_single_line = " ".join(js_ref_lines)

		# Now purify:

		js_ref_as_single_line = re.sub(r'^\s*/\*\s*\$\s*JS PROP REF\s*', '', js_ref_as_single_line)
		js_ref_as_single_line = re.sub(r'\s*\*/', '', js_ref_as_single_line)

		# print(f"as a single line: {js_ref_as_single_line}")

		q_property_js_ref.description.append(js_ref_as_single_line)

		# print(f"After purification of property js ref: {q_property_js_ref.get_description_as_text("  ", 0)}")

	# At this point we need to know if we had a JS REF line or not

	if ref_line_found:
		# We need to read a line
		opening_line = file_handle.readline()

	if "Q_PROPERTY" not in opening_line:
		print("Failed to parse JS REF / Q_PROPERTY line", file=sys.stderr)
		sys.exit(1)

	statement_lines = [ opening_line.replace('\n', '')  ];

	#  If the first line of the Q_PROPERTY statement
	# does not end by ')', then we are dealing with a
	# multiline block.

	if ')' not in opening_line:

		# print("Character ')' is not in the opening_line, this is a multi-line Q_PROPERTY statement")

		while(True):

			line = file_handle.readline()
			line = line.replace('\n', '')
			# print(f"Read follow-up Q_PROPERTY line {line}")
			statement_lines.append(line)

			if ')' in line:
				# print("Character ')' is in the follow-up line")
				# We are done. Exit the loop.
				break;

	else:
		pass
		#print("')' was found in the opening line");

	# At this point we finally encountered the ')', either
	# because it was on the first block opening line or
	# because we iterated in next lines in the file.

	# So now make a single text string out of the string list
	statement_as_single_line = " ".join(statement_lines)

	#print(f"Joined: {statement_as_single_line}")
	simplified_statement_single_line = re.sub(r'\s+', ' ', statement_as_single_line)
	#print(f"Simplified {simplified_statement_single_line}")

	match = property_regex.match(simplified_statement_single_line);

	if(match):

		# print(f"Matched Q_PROPERTY: {match.group(1)} {match.group(2)}")
		q_property_js_ref.statement = f"{match.group(1)} {match.group(2)}"
		# print(f"Returning QpropertyJsRef {str(q_property_js_ref)}")

		return q_property_js_ref

	else:

		print("Failed to extract Q_PROPERTY statement.", file=sys.stderr)
		sys.exit(1)


def parse_q_invokable_block(opening_line, file_handle):

	# opening_line can start these two ways:

	# 1. /*$ JS INVOK REF The .... .... */
	# 2. Q_INVOKABLE explicit Utils(QObject *parent = nullptr);

	q_invokable_js_ref = QinvokableJsRef.QinvokableJsRef()

	ref_line_found = False

	js_ref_lines = [ ]
	invokable_lines = [ ]

	opening_line = re.sub(r'^\s*', '', opening_line)
	opening_line = re.sub(r'\s*\n$', '', opening_line)


	if "$ JS INVOK REF" in opening_line:

		# print(f"JS INVOK REF: {opening_line}")

		ref_line_found = True

		js_ref_lines.append(opening_line)

		if "*/" not in opening_line:

			# print("*/ not in opening_line, multiline text")

			while(True):

				line = file_handle.readline()
				line = line.replace('\n', '')
				# In multiline, there will be chars at the beginning
				# and maybe at the end. Remove them because we need
				# to join the lines laters with ' ' and we do not want
				# multiple spaces in the finished text line.
				line = re.sub(r'^\s*', '', line)
				line = re.sub(r'\s*$', '', line)

				# print(f"Read follow-up JS INVOK REF line {line}")
				js_ref_lines.append(line)

				if "*/" in line:

					# We have finished the JS REF block for the invokable to come.
					break;


		# Join the lines into a single string
		js_ref_as_single_line = " ".join(js_ref_lines)

		# Now purify:

		js_ref_as_single_line = re.sub(r'^\s*/\*\s*\$\s*JS INVOK REF\s*', '', js_ref_as_single_line)
		js_ref_as_single_line = re.sub(r'\s*\*/', '', js_ref_as_single_line)

		# print(f"After purification of invokable js ref: {js_ref_as_single_line}")

		q_invokable_js_ref.description.append(js_ref_as_single_line)

	# At this point we need to know if we had a JS REF line or not

	if ref_line_found:
		# We need to read a line
		opening_line = file_handle.readline()

	if "Q_INVOKABLE" not in opening_line:
		print("Failed to parse JS REF / Q_INVOKABLE line", file=sys.stderr)
		sys.exit(1)

	statement_lines = [ opening_line.replace('\n', '')  ];

	#  If the first line of the Q_INVOKABLE statement
	# does not end by ';', then we are dealing with a
	# multiline block.

	if ';' not in opening_line:

		# print("Character ';' is not in the opening_line, this is a multi-line Q_INVOKABLE statement")

		while(True):

			line = file_handle.readline()
			line = line.replace('\n', '')
			statement_lines.append(line)

			if ';' in line:
				# print("Character ';' is in the follow-up line")
				# We are done. Exit the loop.
				break;

	else:
		# print("';' was found in the opening line");
		pass

	# At this point we finally encountered the ';', either
	# because it was on the first block opening line or
	# because we iterated in next lines in the file.

	# So now make a single text string out of the string list
	statement_as_single_line = " ".join(statement_lines)

	# print(f"Joined: {statement_as_single_line}")
	simplified_statement_single_line = re.sub(r'\s+', ' ', statement_as_single_line)
	# print(f"Simplified {simplified_statement_single_line}")

	match = invokable_regex.match(simplified_statement_single_line);

	if(match):

		# print(f"Matched Q_INVOKABLE: '{match.group(1)}'")
		q_invokable_js_ref.statement = match.group(1)
		# print(f"Returning QinvokableJsRef with {q_invokable_js_ref.statement}")

		return q_invokable_js_ref

	else:
		print("Failed to extract Q_INVOKABLE statement.", file=sys.stderr)
		sys.exit(1)


def parse_description_block(opening_line, file_handle):

	# print(f"Opening line for BEGIN DESCRIPTION")

	description_block_list = [ ];

	# Store all the lines between this opening line and
	# the line that reads "END DESCRIPTION"

	# Pattern matches lines that start with optional whitespace followed by *
	# and some more optional spaces. We want to keep all the spaces to
	# maintain proper formatting.
	remove_star_and_spaces_in_start_pattern = r'^(\s*\*\s*)'

	while(True):

		line = file_handle.readline()

		if not "END DESCRIPTION" in line:

			new_line = re.sub(remove_star_and_spaces_in_start_pattern,
										 r'', line, flags=re.MULTILINE)

			# print(f"Adding new description block line: '{new_line}'")
			description_block_list.append(new_line)
		else:
			break

	# return ("").join(description_block_list)
	return description_block_list


def extract_class_js_reference(header_path):
	"""Returns [ id, properties, invokables]
		 with id = (name_space, class_name)"""

	# print(f"header_path: {header_path}");

	if not os.path.exists(header_path):
		print("File " + header_path + " was not found\n", file = sys.stderr);
		return;

	# We want to iterate in the file  in search for the JS markup of classes
	# that is in the form right before the class declaration:
	#
	# /*  BEGIN JS REFERENCE
	#			namespace: MsXpS::libXpertMassCore
	# 		class name: Polymer
	# */
	#
	# and
	# /*  END JS REFERENCE
	#     namespace: MsXpS::libXpertMassCore
	#     class name: Polymer
	# */
	# right after the class declaration

	# Note that we may have more than one class
	# with JS reference text, so we will return
	# the results as a list.

	parsed_class_js_ref_texts = [ ]

	js_ref_block_opened = False;

	# Closed when first ';' is encountered.
	invokable_block_opened = False;


	# full_js_ref = [ id , properties, invokables ];
	full_js_ref = [ ];

	with open(header_path) as fileHandle:
		line = fileHandle.readline()

		while line:

			if re.match(r'^\s*//', line):
				line = fileHandle.readline()

			# print(f"New line is: {line}");

			if "BEGIN CLASS JS REFERENCE" in line:

				# We are starting a JS reference block

				# Instantiate a class to be used as the
				# container for all the data.
				class_js_ref = ClassJsRef.ClassJsRef()

				js_ref_block_opened = True;

				line = fileHandle.readline()
				# print(f"Parsed line inside BEGIN CLASS JS REFERENCE: {line}")

				# Check if namespace: ... line
				match = name_space_regex.match(line);
				if match:

					if not js_ref_block_opened:
						print("Cannot have name space if JS ref block not opened", file=sys.stderr)
						sys.exit(1)

					name_space = match.group(1);
				else:
					print("Failed to parse the namespace", file=sys.stderr)
					sys.exit(1)

				line = fileHandle.readline()
				# print(f"Parsed line inside BEGIN CLASS JS REFERENCE: {line}")

				# Check if class name: ... line
				match = class_name_regex.match(line);
				if match:

					if not name_space:
						print("Cannot have class name if no name space.", file=sys.stderr)
						sys.exit(1)

					class_name = match.group(1);

					class_js_ref.name_space = name_space
					class_js_ref.class_name = class_name

					# print(f"Could set the name space and class name: {class_js_ref.name_space} - {class_js_ref.class_name}")

				else:
					print("Failed to parse the class name", file=sys.stderr)
					sys.exit(1)


			if "BEGIN DESCRIPTION" in line:
				class_js_ref.description = parse_description_block(line, fileHandle)
				if(len(class_js_ref.description)):
					# print(f"Set description for class {class_js_ref.class_name} is:\n {class_js_ref.description}")
					pass


			if "JS PROP REF" in line or "Q_PROPERTY" in line:

				# print("Detected line with JS PROP REF")

				if not len(class_js_ref.name_space) or not len(class_js_ref.class_name):
					print("Cannot be that a Q_PROPERTY block is opened while there is no name space nor class name.", file=sys.stderr);
					sys.exit(1)

				q_property_js_ref = parse_q_property_block(line, fileHandle)
				q_property_js_ref.name_space = class_js_ref.name_space
				q_property_js_ref.class_name = class_js_ref.class_name

				class_js_ref.properties.append(q_property_js_ref)

				# print(f"Appended new Q_PROPERTY: {q_property_js_ref.statement}")


			if "JS INVOK REF" in line or "Q_INVOKABLE" in line:

				# print("Detected line with JS INVOK REF")

				if not len(class_js_ref.name_space) or not len(class_js_ref.class_name):
					print("Cannot be that a Q_INVOKABLE block is opened while there is no name space nor class name.", file=sys.stderr);
					sys.exit(1)

				q_invokable_js_ref = parse_q_invokable_block(line, fileHandle)
				q_invokable_js_ref.name_space = class_js_ref.name_space
				q_invokable_js_ref.class_name = class_js_ref.class_name

				class_js_ref.invokables.append(q_invokable_js_ref)

				# print(f"Appended new Q_INVOKABLE: {q_invokable_js_ref.statement}")


			if "END CLASS JS REFERENCE" in line:

				if not js_ref_block_opened:
					print("Cannot be that JS ref. block is being closed while it was never opened.", file=sys.stderr)
					sys.exit(1)

				if not len(class_js_ref.name_space) or not len(class_js_ref.class_name):
					print("Cannot be that a Q_PROPERTY block is opened while there is no name space nor class name.", file=sys.stderr);
					sys.exit(1)

				line = fileHandle.readline()
				# print(f"line: {line}")

				# Make sure we are closing the right class JS reference block
				match = name_space_regex.match(line);
				if match:

					end_name_space = match.group(1);
					if end_name_space != name_space:
						print("Cannot be that a JS ref. block is closed with another name space.", file=sys.stderr)
						sys.exit(1)
				else:
					print("The class JS reference block fails to close correctly.", file=sys.stderr)
					sys.exit(1)

				line = fileHandle.readline()
				# print(f"line: {line}")

				match = class_name_regex.match(line);
				if match:

					end_class_name = match.group(1);
					if end_class_name != class_name:
						print("Cannot be that a JS ref. block is closed with another class name.", file=sys.stderr)
						sys.exit(1)

				js_ref_block_opened = False

				# So we've done parsing one class JS reference text, add the instantiated
				# class to the returned list.

				parsed_class_js_ref_texts.append(class_js_ref)

			# Go on with potential new class with JS reference text.
			line = fileHandle.readline()

		# print(f"Finished parsing {header_path}\n")

	return parsed_class_js_ref_texts


def listAllJsRefFiles(root_dir, js_ref_bock_tag):

		root_path = Path(root_dir)
		matching_files = []


		# Recursively search for both .hpp and .h files
		hpp_files = root_path.rglob("*.hpp")
		h_files = root_path.rglob("*.h")

		# Combine both iterators
		all_header_files = itertools.chain(hpp_files, h_files)

		for header_file in all_header_files:

				try:
						with open(header_file, 'r', encoding='utf-8') as f:
								content = f.read()
								if js_ref_bock_tag in content:
										matching_files.append(str(header_file))

				except (UnicodeDecodeError, PermissionError, OSError) as e:
						print(f"Could not read {header_file}: {e}", file=sys.stderr)

		return matching_files
